// 
// TraceHook - Application designed to monitor processes marked
//             as malware together with their children (recursively) and
//             dumping their memory contents after a specified interval or
//             Ctrl+C signal.
// 
// Author:  j00ru//vx
// Date:    2009-08-04
// Version: 0.0.2
// Module:  User-mode client
//
#define _WIN32_WINNT 0x0600
#include <cstdio>
#include <cstdlib>
#include <string>
#include <vector>
#include <windows.h>
#include <ddk\ntapi.h>
#include <psapi.h>
#include <shlwapi.h>
#include <dbghelp.h>
#include "driver.h"
using namespace std;

#define INFO(text) {\
                     printf("[INFO]  ");\
                     printf text;\
                     fflush(stdout);\
                   }
#define ERR(text)  {\
                     printf("[ERROR] ");\
                     printf text;\
                     fflush(stdout);\
                   }
#define DIE(text)  {\
                     printf text;\
                     ExitProcess(1);\
                     fflush(stdout);\
                   }
#define SUCCESS    {\
                      puts("Success!");\
                      fflush(stdout);\
                   }




/* The TraceHook device name in a short form
 */
const CHAR ServiceName[]    = "TraceHook";
const CHAR DeviceName[]     = "\\\\.\\TraceHook";
HANDLE hDevice;



/* Kernel-mode module communication routines 
 */
BOOL QueryDriverVersion(HANDLE hDriver);
BOOL SendFinishSignal(HANDLE hDriver);
BOOL AddProcessToList(HANDLE hDriver,DWORD dwProcessId);
BOOL RemoveProcessFromList(HANDLE hDriver,DWORD dwProcessId);
BOOL SendDumpPath(HANDLE hDriver,PCHAR DumpPath);
BOOL SetDumpSettings(HANDLE hDriver, DumpMode *DumpSettings);

/* Kernel-module loading routines
 */
BOOL LoadDriver();
VOID UnloadDriver();



/* LoadDriver routine
 *
 * Arguments:
 *  None
 * 
 * Return value:
 *  TRUE, if the driver has been successfully loaded and executed
 *  FALSE otherwise
 */
BOOL LoadDriver()
{
  SC_HANDLE hSCManager;
  SC_HANDLE hService;
  SERVICE_STATUS ss;
  CHAR Path[4096];


  /* Before trying to manually load the driver, make sure it
   * is not currently loaded.
   */
  UnloadDriver();

  /* Form the initial version of driver path
   */
  GetCurrentDirectory(sizeof(Path),Path);
  strcat(Path, "\\");
  strcat(Path, DriverFileName);

  /* Open the SC manager
   */
  hSCManager = OpenSCManager(NULL, NULL, SC_MANAGER_CREATE_SERVICE);
  if(!hSCManager)
  {
    DIE(("OpenSCManager() failed\r\n"));
    return false;
  }

  /* Check if the service isn't already active
   */
  hService = OpenService(hSCManager, ServiceName, SERVICE_START | DELETE | SERVICE_STOP | SERVICE_QUERY_STATUS);
  if(!hService)
  {
    /* Try to create our service
     */
    hService = CreateService(hSCManager, ServiceName, ServiceName,
                            	SERVICE_START|DELETE|SERVICE_STOP|SERVICE_QUERY_STATUS, SERVICE_KERNEL_DRIVER,
	                            SERVICE_DEMAND_START, SERVICE_ERROR_IGNORE, Path,
                             	NULL, NULL, NULL, NULL, NULL);
    if(!hService)
    {
      DIE(("CreateService() and OpenService() failed\r\n"));
      CloseServiceHandle(hSCManager);
      return FALSE;
    }
  }

  /* If the service is not already running, try to make it run
   */
  QueryServiceStatus(hService, &ss);
  if(ss.dwCurrentState != SERVICE_RUNNING)
  {
    if(!StartService(hService, 0, NULL))
    {
      DIE(("StartService() failed. LastError: 0x%.8x\r\n",(UINT)GetLastError()));
      ControlService(hService, SERVICE_CONTROL_STOP, &ss);
      DeleteService(hService);
      CloseServiceHandle(hService);
      CloseServiceHandle(hSCManager);
      return FALSE;
    }
  }

  CloseServiceHandle(hService);
  CloseServiceHandle(hSCManager);
  return TRUE;
}


/* UnloadDriver routine
 *
 * Arguments:
 *  None
 * 
 * Return value:
 *  None
 */
VOID UnloadDriver()
{
  SC_HANDLE hSCManager;
  SC_HANDLE hService;
  SERVICE_STATUS ss;

  /* Open the service manager
   */
  hSCManager = OpenSCManager(NULL, NULL, SC_MANAGER_CREATE_SERVICE);

  /* Open the service
   */
  hService = OpenService(hSCManager, ServiceName, SERVICE_START | DELETE | SERVICE_STOP);

  /* Stop and delete the service
   */
  ControlService(hService, SERVICE_CONTROL_STOP, &ss);
  DeleteService(hService);

  /* Close handles
   */
  CloseServiceHandle(hService);
  CloseServiceHandle(hSCManager);
}



/* MakeFinalDump routine
 *
 * Arguments:
 *  None
 * 
 * Return value:
 *  None
 *
 * The function is responsible for signalising the driver that the 
 * client application is being closed, thus forcing it to terminate
 * every process inside the malware list (and therefore dumping them).
 * After that, the driver gets unloaded from the operating system.
 */
void MakeFinalDump()
{
  INFO(("Sending a FINISH signal to the driver...\r\n"));
  SendFinishSignal(hDevice);

  INFO(("Unloading the driver from operating system.\r\n"));
  UnloadDriver();

  INFO(("Good bye."));
  ExitProcess(0);
}


/* QueryDriverVersion routine
 *
 * Arguments:
 *  hDriver - handle to the active TraceHook driver
 * 
 * Return value:
 *  TRUE, if the version has been successfully received
 *  FALSE otherwise
 */
BOOL QueryDriverVersion(HANDLE hDriver)
{
  static char VersionBuffer[4096];
  DWORD BytesReturned;
  
  /* Send an IOCTL signal to the driver and wait for a response
   */
  if(!DeviceIoControl(hDriver,
                      IOCTL_QUERY_VERSION,
                      NULL,
                      0,
                      VersionBuffer,
                      sizeof(VersionBuffer),
                      &BytesReturned,
                      NULL))
  {
    ERR(("DeviceIoControl() failed. LastError: 0x%.8x\n",(UINT)GetLastError()));
    return FALSE;
  }

  /* Print the obtained string
   */
  INFO(("Driver Version: %s\n",VersionBuffer));
  return TRUE;
}


/* SendFinishSignal routine
 *
 * Arguments:
 *  hDriver - handle to the active TraceHook driver
 * 
 * Return value:
 *  TRUE, if all the monitored processes have been succesfully dumped
 *  FALSE otherwise
 */
BOOL SendFinishSignal(HANDLE hDriver)
{
  DWORD BytesReturned;

  /* Send an IOCTL signal to the driver and wait for a response
   */
  if(!DeviceIoControl(hDriver,
                      IOCTL_FINISH,
                      NULL,
                      0,
                      NULL,
                      0,
                      &BytesReturned,
                      NULL))
  {
    ERR(("DeviceIoControl() failed. LastError: 0x%.8x\n",(UINT)GetLastError()));
    return FALSE;
  }

  return TRUE;  
}


/* AddProcessToList routine
 *
 * Arguments:
 *  dwProcessId - PID of the process to add to the tracing list
 * 
 * Return value:
 *  TRUE, if the signal has been successfully handled by the driver
 *  FALSE otherwise
 */
BOOL AddProcessToList(HANDLE hDriver,DWORD dwProcessId)
{
  DWORD BytesReturned;

  /* Send an IOCTL signal to the driver and wait for a response
   */
  if(!DeviceIoControl(hDriver,
                      IOCTL_ADD_PROCESS,
                      &dwProcessId,
                      sizeof(DWORD),
                      NULL,
                      0,
                      &BytesReturned,
                      NULL))
  {
    ERR(("DeviceIoControl() failed. LastError: 0x%.8x\n",(UINT)GetLastError()));
    return FALSE;
  }

  return TRUE;
}


/* RemoveProcessFromList routine
 *
 * Arguments:
 *  dwProcessId - PID of the process to remove from the tracing list
 * 
 * Return value:
 *  TRUE, if the signal has been successfully handled by the driver
 *  FALSE otherwise
 */
BOOL RemoveProcessFromList(HANDLE hDriver,DWORD dwProcessId)
{
  DWORD BytesReturned;

  /* Send an IOCTL signal to the driver and wait for a response
   */
  if(!DeviceIoControl(hDriver,
                      IOCTL_REMOVE_PROCESS,
                      &dwProcessId,
                      sizeof(DWORD),
                      NULL,
                      0,
                      &BytesReturned,
                      NULL))
  {
    ERR(("DeviceIoControl() failed. LastError: 0x%.8x\n",(UINT)GetLastError()));
    return FALSE;
  }

  return TRUE;
}



/* SendDumpPath routine
 *
 * Arguments:
 *  DumpPath - New path to the directory, where all process dumps are stored in.
 * 
 * Return value:
 *  TRUE, if the signal has been successfully handled by the driver
 *  FALSE otherwise
 */
BOOL SendDumpPath(HANDLE hDriver,PWCHAR DumpPath)
{
  DWORD BytesReturned;
  
  /* Send an IOCTL signal to the driver and wait for a response
   */
  if(!DeviceIoControl(hDriver,
                      IOCTL_SET_DUMP_PATH,
                      DumpPath,
                      wcslen(DumpPath)*sizeof(WCHAR),
                      NULL,
                      0,
                      &BytesReturned,
                      NULL))
  {
    ERR(("DeviceIoControl() failed. LastError: 0x%.8x\n",(UINT)GetLastError()));
    return FALSE;
  }

  return TRUE;
}


/* SetDumpSettings routine
 *
 * Arguments:
 *  DumpSettings - pointer to a structure describing the desired dumping settings
 * 
 * Return value:
 *  TRUE, if the signal has been successfully handled by the driver
 *  FALSE otherwise
 */
BOOL SetDumpSettings(HANDLE hDriver, DumpMode *DumpSettings)
{
  DWORD BytesReturned;

  /* Send an IOCTL signal to the driver and wait for a response
   */
  if(!DeviceIoControl(hDriver,
                      IOCTL_SET_DUMP_MODE,
                      DumpSettings,
                      sizeof(DumpMode),
                      NULL,
                      0,
                      &BytesReturned,
                      NULL))
  {
    ERR(("DeviceIoControl() failed. LastError: 0x%.8x\n",(UINT)GetLastError()));
    return FALSE;
  }

  return TRUE;
}




/* WaitForUserAction routine
 *
 * Arguments:
 *  Not important
 * 
 * Return value:
 *  Not important - always 0
 */
DWORD WaitForUserAction(LPVOID lpParameter)
{
  /* Wait about 0.5 second to let the main() function initialize the necessary environment
   */
  Sleep(500);

  INFO(("TraceHook module is active, monitoring the malware processes.\r\n\r\n"));
  INFO(("Press ENTER to stop the analysis.\r\n"));

  /* Wait for the user to press ENTER
   */
  getchar();

  return 0;
}



/* ApcThreadRoutine routine
 *
 * Arguments:
 *  Unused
 * 
 * Return value:
 *  Unused
 *
 *  This routine is used by the TraceHook.EXE process as an "APC container" -
 *  a thread in which all the kernel->user APC routines are being executed.
 */
DWORD ApcThreadRoutine(LPVOID lpParameter)
{
  while(1) SleepEx(INFINITE,TRUE);
}



/* ApcMiniDumpRoutine routine
 *
 * Arguments:
 *  NormalContext   - contains the ID of the process to be dumped
 *  SystemArgument1 - contains a handle to an Event object
 *  SystemArgument2 - contains a handle to the dump file being written to
 *
 * Return value:
 *  None
 */
VOID __stdcall ApcMiniDumpRoutine(
    PVOID NormalContext,
    PVOID SystemArgument1,
    PVOID SystemArgument2)
{
  DWORD  dwProcessId  = (DWORD)NormalContext;
  HANDLE hRemoteEvent = (HANDLE)SystemArgument1;
  HANDLE hRemoteDump  = (HANDLE)SystemArgument2;
  HANDLE hProcess,hEvent,hDumpFile;

  INFO(("ApcMiniDumpRoutine(0x%.8x) called.\r\n",(UINT)dwProcessId));

  /* First of all, let's try to open the desired process
   */ 
  hProcess = OpenProcess(PROCESS_ALL_ACCESS,FALSE,dwProcessId);
  if(!hProcess)
  {
    ERR(("OpenProcess failed. LastError: 0x%.8x\r\n",(UINT)GetLastError()));
    return;
  }

  /* Duplicate the hEvent handle
   */
  if(!DuplicateHandle(
        hProcess,
        hRemoteEvent,
        GetCurrentProcess(),
        &hEvent,
        0,
        FALSE,
        DUPLICATE_SAME_ACCESS))
  {
    ERR(("DuplicateHandle failed. LastError: 0x%.8x\r\n",(UINT)GetLastError()));
    CloseHandle(hProcess);
    return;
  }

  /* Do the same with hDumpFile
   */
  if(!DuplicateHandle(
        hProcess,
        hRemoteDump,
        GetCurrentProcess(),
        &hDumpFile,
        0,
        FALSE,
        DUPLICATE_SAME_ACCESS))
  {
    ERR(("DuplicateHandle failed. LastError: 0x%.8x\r\n",(UINT)GetLastError()));
    SetEvent(hEvent);    
    CloseHandle(hProcess);
    CloseHandle(hEvent);
    return;    
  }
  
  /* Perform the minidump itself, based on handles passed by the device driver
   */
  if(!MiniDumpWriteDump(
        hProcess,
        dwProcessId,
        hDumpFile,
        (MINIDUMP_TYPE)(
          (UINT)MiniDumpWithDataSegs|
          (UINT)MiniDumpWithFullMemory|
          (UINT)MiniDumpWithHandleData|
          (UINT)MiniDumpWithProcessThreadData),
        NULL,
        NULL,
        NULL))
  {
    ERR(("MiniDumpWithFullMemory failed. LastError: 0x%.8x\r\n",(UINT)GetLastError()));
    SetEvent(hEvent);
    CloseHandle(hProcess);
    CloseHandle(hEvent);
    CloseHandle(hDumpFile);
    return;
  }

  /* Set the hEvent status as Signalised and close all the pending handles
   */
  SetEvent(hEvent);
  CloseHandle(hEvent);  
  CloseHandle(hProcess);
  CloseHandle(hDumpFile);

  INFO(("Process 0x%.8x successfully dumped to disk.\r\n",(UINT)dwProcessId));
}



/* AnsiToUnicode routine
 *
 * Arguments:
 *  AnsiString - pointer to the string to convert
 * 
 * Return value:
 *  Pointer to an UNICODE-encoded version of text passed in the parameter
 */
PWCHAR AnsiToUnicode(PCHAR AnsiString)
{
  UINT Length=strlen(AnsiString);
  PWCHAR NewString = new WCHAR[Length+1];

  MultiByteToWideChar(CP_ACP,0,AnsiString,Length,NewString,Length);
  NewString[Length] = L'\0';

  return NewString;
}


/* ShowHelp routine
 *
 * Arguments:
 *  None
 * 
 * Return value:
 *  None
 */
void ShowHelp()
{
  DIE(( " Usage: TraceHook.exe <target executable> [options]\r\n"
        "\r\n"
        " Available options:\r\n"
        "  -pPATH       Sets the dump destination directory to PATH.\r\n"
        "               The default dump path is C:\\dump.\r\n"
        "\r\n"
        "  -iINTERVAL   Turns the counter mode on. Makes the application terminate\r\n"
        "               and automatically dump all the monitored processes after the\r\n"
        "               specified amount of time.\r\n"
        "\r\n"
        "               Note: The INTERVAL value is the number of seconds to wait till\r\n"
        "                     dumping all processes.\r\n"
        "\r\n"
        "  -m           Activates the MiniDump mode.\r\n"
        "\r\n"
        "  -h           Displays this message.\r\n"));
}




/* main routine
 *
 * Arguments:
 *  Not important (information about the console arguments)
 * 
 * Return value:
 *  Not important
 */
int main(int argc, char** argv)
{
  PCHAR ActiveDumpPath = (PCHAR)UserDumpPath;  
  UNICODE_STRING KernelDumpPath;
  DWORD WaitInterval = 0;
  HANDLE hThread;
  BOOL MiniDumpMode=FALSE;
  static DumpMode DumpSettings;

  puts("-=*( TraceHook " TRACEHOOK_VERSION " by j00ru//vx )*=-");

  if(argc<2)
  {
    ShowHelp();
    return 1;   // should not even get here
  }

  INFO(("Parsing parameters... \r\n"));
  for( int i=2;i<argc;i++ )
  {
    if(argv[i][0]!='-')
      DIE(("Invalid %d argument!",i+1));
    
    switch(argv[i][1])
    {
      case 'p':
        ActiveDumpPath = &argv[i][2];
        break;
      case 'i':
        WaitInterval = atoi(&argv[i][2])*1000;
        break;
      case 'm':
        MiniDumpMode = TRUE;
        break;
      case 'h':
        ShowHelp();
        break;
      default:
        DIE(("Invalid option in %d argument!",i+1));
    }
  }

  INFO(("Loading driver... "));
  if(!LoadDriver())
    DIE(("Failed"));
  SUCCESS;

  INFO(("Opening driver... "));
  hDevice = CreateFileA(DeviceName,GENERIC_READ|GENERIC_WRITE,0,NULL,OPEN_EXISTING,0,NULL);
  if(hDevice==NULL)
  {
    UnloadDriver();
    DIE(("Failed"));
  }
  SUCCESS;  

  INFO(("Querying for the driver version...\r\n"));
  if(!QueryDriverVersion(hDevice))
  {
    UnloadDriver();
    DIE(("Failed"));
  }

  INFO(("Creating the %s directory...",ActiveDumpPath));
  if(!CreateDirectory(ActiveDumpPath,NULL))
  {
    if(GetLastError()!=ERROR_ALREADY_EXISTS)
    {
      UnloadDriver();
      DIE(("Failed"));
    }
  }
  SUCCESS;

  /* Convert the ANSI DosPath to internal NT relative path
   */
  RtlDosPathNameToNtPathName_U(AnsiToUnicode(ActiveDumpPath),&KernelDumpPath,NULL,NULL);

  INFO(("Sending the dump directory path to TraceHook.sys..."));
  if(!SendDumpPath(hDevice,(PWCHAR)KernelDumpPath.Buffer))
  {
    UnloadDriver();
    DIE(("Failed"));
  }
  SUCCESS;

  INFO(("Setting the requested dump mode..."));
  if(MiniDumpMode)
  {
    HANDLE hThreadApc;

    /* Create a special service thread, staying in the Alertable State all the time.
     * It is used by the device driver to queue special APCs to dump the specified process
     * memory from user-mode
     */
    hThreadApc = CreateThread(NULL,0,(LPTHREAD_START_ROUTINE)ApcThreadRoutine,NULL,0,NULL);
    if(hThreadApc==NULL)
    {
      UnloadDriver();
      DIE(("Failed"));
    }

    DumpSettings.DumpType         = MiniDump;
    DumpSettings.dwTraceHookPid   = GetCurrentProcessId();
    DumpSettings.hThreadApc       = hThreadApc;
    DumpSettings.ApcRoutine       = (PIO_APC_ROUTINE)ApcMiniDumpRoutine;

    if(!SetDumpSettings(hDevice,&DumpSettings))
    {
      UnloadDriver();
      DIE(("Failed"));
    }
  }
  else
  {
    DumpSettings.DumpType       = FullMemoryDump;
    DumpSettings.dwTraceHookPid = 0;
    DumpSettings.hThreadApc     = NULL;
    DumpSettings.ApcRoutine     = NULL;
  
    if(!SetDumpSettings(hDevice,&DumpSettings))
    {
      UnloadDriver();
      DIE(("Failed"));
    }
  }
  SUCCESS;


  /* Create the specified process
   */
  STARTUPINFO si;
  PROCESS_INFORMATION pi;
  memset(&si,0,sizeof(si));
  memset(&pi,0,sizeof(pi));
  si.cb = sizeof(si);

  INFO(("Creating a suspended process..."));
  if(!CreateProcess(NULL,argv[1],NULL,NULL,false,CREATE_NEW_CONSOLE|CREATE_SUSPENDED,NULL,NULL,&si,&pi))
  {
    RemoveProcessFromList(hDevice,GetCurrentProcessId());
    UnloadDriver();
    DIE(("Failed"));
  }
  SUCCESS;

  INFO(("Adding the %u ProcessId to the monitor list...\r\n",(UINT)pi.dwProcessId));
  AddProcessToList(hDevice,pi.dwProcessId);
  
  INFO(("Resuming the target process..."));
  if(!ResumeThread(pi.hThread))
  {
    UnloadDriver();
    DIE(("Failed"));
  }
  SUCCESS;

  INFO(("Creating a thread waiting for user interaction..."));
  if((hThread=CreateThread(NULL,0,(LPTHREAD_START_ROUTINE)WaitForUserAction,NULL,0,NULL))==NULL)
  {
    UnloadDriver();
    DIE(("Unable to create a local thread! :o\r\n"));
  }
  SUCCESS;

  /* Depending on the arguments, wait forever (till the user presses ENTER)
   * or for the specified number of seconds.
   */
  if(WaitInterval!=0)
    WaitForSingleObject(hThread,WaitInterval);
  else
    WaitForSingleObject(hThread,INFINITE);


  /* Inform the driver about the process termination, so as to let it 
   * dump all the remaining processes - after it is done, the driver
   * gets unloaded.
   */
  MakeFinalDump();

  /* We should not get ths far, at all.
   */
  while(1) Sleep(1);
  return 0;
}

