// 
// TraceHook - Application designed to monitor processes marked
//             as malware together with their children (recursively) and
//             dumping their memory contents after a specified interval or
//             Ctrl+C signal.
// 
// Author:  j00ru//vx
// Date:    2009-08-04
// Version: 0.0.2
// Module:  Kernel-mode device driver
//
// Special thanks to Cyclops for pointing out a DoS vulnerability. The issue
// has been solved by restricting access to IOCTL commands - processes marked as
// malware are not able to directly communicate with the driver, now.
//
#include <stdio.h>
#include <ntddk.h>
#include <windef.h>
#include <ntdef.h>
#include "krnl_def.h"
#include "driver.h"
#define PROCESS_TERMINATE	1
#define THREAD_SET_CONTEXT	16


/*
 * Internal driver variables
 */
PDEVICE_OBJECT pDeviceObject = NULL;  // Just a device object associated with the driver
KSPIN_LOCK SpLock;                    // Global spinlock structure
UINT NameOffset;                      // EPROCESS structure offset, where the 16-byte-long process name is stored
DumpMode DumpSettings;                // Some settings, affecting the mechanism used to dump the process memory


/* A very simple structure describing the list of PIDs being currently monitored
 */
struct _ProcessList
{
  UINT Count;
  UINT Size;
  HANDLE* List;
} ProcessList;

/* Array containing the raw HDD path, where the process memory dumps are written to.
 */
WCHAR GlobalDumpPath[MAX_PATH+1] = L"\\\\??\\C\\dump";

/* Current TraceHook version string.
 * Passed back to user-mode in the QueryVersion IOCTL call
 */
const  CHAR DriverVersion[] =  "TraceHook driver " TRACEHOOK_VERSION " by j00ru//vx";



/* Internal functions' definitions
 */
VOID DriverUnload(PDRIVER_OBJECT pDriverObject);
NTSTATUS DriverCmd(PDEVICE_OBJECT, PIRP);
NTSTATUS DriverUnsupported(PDEVICE_OBJECT, PIRP);



/* GetLocationOfProcessName routine
 *
 * Arguments:
 *  None
 * 
 * Return value:
 *  Offset from the beginning of the EPROCESS structure, at which 
 *  the current process name remains.
 *
 *  This code has been copied from the "Subverting Windows Kernel" book.
 */
ULONG GetLocationOfProcessName()
{
   ULONG ul_offset;
   PEPROCESS CurrentProc = PsGetCurrentProcess();
   // This will fail if the EPROCESS grows larger
   // than a page size.
   for(ul_offset = 0; ul_offset < PAGE_SIZE; ul_offset++)
   {
      if( !strncmp( "System", (PCHAR) CurrentProc + ul_offset,
                    strlen("System")))
      {
         return ul_offset;
      }
   }
   return (ULONG)0;
}



/* AddProcessToList routine
 *
 * Arguments:
 *  ProcessId - A single 32-bit PID value of the process to add
 * 
 * Return value: None
 */
void AddProcessToList(HANDLE ProcessId)
{
  KIRQL OldIrql;

  /* Synchronize access to the list (one function at a time)
   */
  KeAcquireSpinLock(&SpLock,&OldIrql);

  /* In case the array size has been exceeded, reallocate teh buffer
   * by allocating a bigger one and copying the old piece of data there
   */
  if(ProcessList.Count>=ProcessList.Size || ProcessList.List==NULL)
  {
    HANDLE* OldList = ProcessList.List;

    /* Widen the current buffer by 4096 bytes
     */
    ProcessList.Size += 4096;
    ProcessList.List = ExAllocatePool(NonPagedPool,ProcessList.Size*sizeof(HANDLE));

    /* Copy old memory and release old pool memory
     */
    RtlMoveMemory(ProcessList.List,OldList,ProcessList.Count*sizeof(HANDLE));
    ExFreePool(OldList);
  }

  /* Put the new item into the list
   */
  ProcessList.List[ProcessList.Count++] = ProcessId;  

  /* Let other functions access the list
   */
  KeReleaseSpinLock(&SpLock,OldIrql);
}


/* RemoveProcessFromList routine
 *
 * Arguments:
 *  ProcessId - A single 32-bit PID value of the process to remove
 * 
 * Return value: None
 */
BOOL RemoveProcessFromList(HANDLE ProcessId)
{
  KIRQL OldIrql;
  UINT i;
  BOOL Result=FALSE;

  /* Synchronize access to the list (one function at a time)
   */
  KeAcquireSpinLock(&SpLock,&OldIrql);

  /* Iterate thru all the list items
   */
  for( i=0;ProcessList.List!=NULL && i<ProcessList.Count;i++ )
  {
    /* If the right PID is found, remove it by moving the following items 4 bytes back
     */
    if(ProcessList.List[i]==ProcessId)
    {
      RtlMoveMemory(&ProcessList.List[i],&ProcessList.List[i+1],(ProcessList.Size-i-1)*sizeof(HANDLE));
      ProcessList.Count--;
      Result=TRUE;
      break;
    }
  }

  /* Let other functions access the list
   */
  KeReleaseSpinLock(&SpLock,OldIrql);
  return Result;
}


/* IsProcessInList routine
 *
 * Arguments:
 *  ProcessId - A single 32-bit PID value of the process to search for
 * 
 * Return value:
 *  - TRUE, if the specified process is found inside the list
 *  - FALSE otherwise
 */
BOOL IsProcessInList(HANDLE ProcessId)
{
  KIRQL OldIrql;
  UINT i;
  BOOL Result=FALSE;

  /* Synchronize access to the list (one function at a time)
   */
  KeAcquireSpinLock(&SpLock,&OldIrql);

  /* Iterate thru all the list items
   */
  for( i=0;ProcessList.List!=NULL && i<ProcessList.Count;i++ )
  {
    if(ProcessList.List[i]==ProcessId)
    {
      Result=TRUE;
      break;
    }
  }

  /* Let other functions access the list
   */
  KeReleaseSpinLock(&SpLock,OldIrql);  
  return Result;
}


/* ApcKernelRoutine routine
 *
 * Arguments:
 *  Unused
 * 
 * Return value:
 *  None
 */
VOID
ApcKernelRoutine(
    IN struct _KAPC *Apc,
    IN OUT PKNORMAL_ROUTINE *NormalRoutine,
    IN OUT PVOID *NormalContext,
    IN OUT PVOID *SystemArgument1,
    IN OUT PVOID *SystemArgument2
    )
{
  /* XXX: do we have to write anything here?
   */
}



/* DumpProcessMemory routine
 *
 * Arguments:
 *  dwProcessId - PID of the process to be dumped. This value is passed back
 *                to a user-mode routine, when creating a minidump file.
 * 
 * Return value:
 *  None
 *
 * The routine dumps the contents of user-mode virtual memory into a new
 * file, inside the DumpPath[] directory, named in the below format:
 *  %ProcessId%.%ProcessName%.dmp
 */
void DumpProcessMemory(HANDLE dwProcessId)
{
  NTSTATUS NtStatus;
  OBJECT_ATTRIBUTES ObjectAttributes;
  IO_STATUS_BLOCK IoStatusBlock;
  HANDLE FileHandle;
  UINT i;
  CHAR Path[1024];
  ANSI_STRING    aPath;
  UNICODE_STRING uPath;
  static BYTE Buffer[4096];

  
  /* Create the desired file (with a properly formatted filename)
   */
  sprintf(Path,"%ws\\%.4x.%s.dmp",GlobalDumpPath,PsGetCurrentProcessId(),
                                  (PCHAR)((PCHAR)PsGetCurrentProcess()+NameOffset));

  /* Convert the ANSI Path to UNICODE wPath
   */
  RtlInitAnsiString(&aPath,Path);
  RtlAnsiStringToUnicodeString(&uPath,&aPath,TRUE);
  
  InitializeObjectAttributes(&ObjectAttributes,
      &uPath,
      OBJ_CASE_INSENSITIVE|OBJ_OPENIF,
      NULL,NULL);

  /* Open the desired dump file, no matter what dumping method we're going to use
   */
  if(!NT_SUCCESS(ZwCreateFile(&FileHandle,GENERIC_WRITE,&ObjectAttributes,&IoStatusBlock,NULL,
                FILE_ATTRIBUTE_NORMAL,0,FILE_CREATE,FILE_NON_DIRECTORY_FILE|FILE_SYNCHRONOUS_IO_ALERT,NULL,0)))
  {
    DbgPrint("[-] TraceHook.SYS: ZwCreateFile(%s); failed.\r\n",Path);
    return;
  }
  DbgPrint("[+] TraceHook.SYS: Dump file successfully created.\r\n",Path);


  switch(DumpSettings.DumpType)
  {
    case FullMemoryDump:
    {
      /* Loop thru all the possible 32-bit user-mode addresses
       */
      for( i=0x00000000;i<0x80000000;i+=0x1000 )
      {
        __try
        {
          /* Try to read (move) the current memory page into a kernel-mode buffer
           */
          RtlCopyMemory(Buffer,(PVOID)i,0x1000);

          /* If we're still here, the address must be valid - dump it to the file
          */
          if(!NT_SUCCESS(ZwWriteFile(FileHandle,NULL,NULL,NULL,&IoStatusBlock,Buffer,sizeof(Buffer),NULL,NULL)))
          {
            /* Writing the page contents failed.
             * XXX: Do we need to do anything in such case?
             */
          }
        }
        __except(EXCEPTION_EXECUTE_HANDLER)
        {
        }
      }
    }
    break;

    case MiniDump:
    {
      PKAPC DumpApc;
      PKTHREAD DumpApcThread;
      HANDLE hTraceHook;
      HANDLE hValidApcThread;
      HANDLE hEvent;
      CLIENT_ID Cid = {(HANDLE)DumpSettings.dwTraceHookPid,0};
      OBJECT_ATTRIBUTES ObjectAttributes;
      LARGE_INTEGER Interval;


      /* Create a special Event object, in order to wait for the user-mode APC routine
       * to complete the minidump writing before letting the process terminate.
       */
      if(!NT_SUCCESS(ZwCreateEvent(
              &hEvent,
              EVENT_ALL_ACCESS,
              NULL,
              SynchronizationEvent,
              FALSE)))
      {
        DbgPrint("[-] TraceHook.SYS: ZwCreateEvent failed.\r\n");
        return;
      }

      /* Initialize the structure with NULLs
       */
      InitializeObjectAttributes(&ObjectAttributes,NULL,0,NULL,NULL);        

      /* First of all, open a HANDLE to the TraceHook.EXE process
       */
      if(!NT_SUCCESS(ZwOpenProcess(
                      &hTraceHook,
                      PROCESS_DUP_HANDLE,
                      &ObjectAttributes,
                      &Cid)))
      {
        DbgPrint("[-] TraceHook.SYS: ZwOpenProcess failed.\r\n");
        ZwClose(hEvent);
        return;
      }

      if(!NT_SUCCESS(ZwDuplicateObject(
                      hTraceHook,
                      DumpSettings.hThreadApc,
                      NtCurrentProcess(),
                      &hValidApcThread,
                      0,
                      FALSE,
                      DUPLICATE_SAME_ACCESS)))
      {
        DbgPrint("[-] TraceHook.SYS: ZwDuplicateObject failed.\r\n");
        ZwClose(hEvent);
        ZwClose(hTraceHook);
        return;
      }

      /* Close the old TraceHook.EXE process handle
       */
      ZwClose(hTraceHook);      

      /* Translate a valid hThread HANDLE into a KTHREAD structure pointer
       */
      if(!NT_SUCCESS(ObReferenceObjectByHandle(
                        hValidApcThread,
                        THREAD_SET_CONTEXT,
                        NULL,
                        ExGetPreviousMode(),
                        &DumpApcThread,
                        NULL)))
      {
        DbgPrint("[-] TraceHook.SYS: ObReferenceObjectByHandle failed.\r\n");
        ZwClose(hEvent);
        ZwClose(hValidApcThread);
        return;
      }


      /* Allocate some memory for the DumpApc structure
       */
      DumpApc = ExAllocatePool(NonPagedPool,sizeof(KAPC));
      if(!DumpApc)
      {
        DbgPrint("[-] TraceHook.SYS: ExAllocatePool failed.\r\n");
        ZwClose(hEvent);
        return;
      }

      /* Initialize the internal KAPC structure with all the values obtained before
       */
      KeInitializeApc(DumpApc,DumpApcThread,OriginalApcEnvironment,
                      ApcKernelRoutine,NULL,(PKNORMAL_ROUTINE)DumpSettings.ApcRoutine,UserMode,(PVOID)dwProcessId);

      /* Insert the KAPC structure into the thread's APC queue
       */
      if(!KeInsertQueueApc(DumpApc,(PVOID)hEvent,(PVOID)FileHandle,0))
      {
        DbgPrint("[-] TraceHook.SYS: KeInsertQueueApc failed.\r\n");
        ZwClose(hEvent);
        ZwClose(hValidApcThread);
        ExFreePool(DumpApc);
        return;
      }

      /* Wow, the APC is probably executing now!
       * Let's wait for the routine to complete before letting the process completely terminate.
       *
       * The current (constant) interval time is 5 minutes.
       */
      Interval.QuadPart = -(5LL*60LL * 10000000LL);
      ZwWaitForSingleObject(hEvent,FALSE,&Interval);
 
      ZwClose(hEvent);
      ZwClose(hValidApcThread);
    }
    break;

    default:
      break;
  }

  /* Close the file handle
   */
  ZwClose(FileHandle);  

  DbgPrint("[+] TraceHook.SYS: Dumping done, exiting.\r\n");
}


/* NotifyRoutine routine
 *
 * Arguments:
 *  ParentId  - PID of the parent process
 *  ProcessId - PID of the current process
 *  Create    - Information about whether the notify concerns a process creation or
 *              destruction.
 * 
 * Return value:
 *  None
 *
 * This callback routine is being executed everytime a process is being created/removed
 * from the system. This gives us the opportunity to control these events and perform
 * appropriate actions (like dumping the image memory right before the process gets 
 * completely destroyed).
 */
VOID
NotifyRoutine(
    IN HANDLE  ParentId,
    IN HANDLE  ProcessId,
    IN BOOLEAN Create
    )
{
  DbgPrint("[+] TraceHook.SYS: NotifyRoutine() called:\r\n"
           "     ParentId:  0x%.8x\r\n"
           "     ProcessId: 0x%.8x\r\n"
           "     Create:    %s\r\n",ParentId,ProcessId,Create?"TRUE":"FALSE");

  if(Create)
  {
    /* If the parent's PID is found inside the list, do the same with the current process
     */
    if(IsProcessInList(ParentId))
    {
      DbgPrint("[+] TraceHook.SYS: Adding the 0x%.8x process to malware list.\r\n",ProcessId);
      AddProcessToList(ProcessId);
    }
  }
  else
  {
    /* Check if the process is one of the monitored ones
     */
    if(IsProcessInList(ProcessId))
    {
      /* XXX: The code is taking advantage of the fact that in case of process termination,
       *      the notify routine is being called in the context of the destination process, 
       *      thus we are able to access and dump its memory by simply accessing it's virtual
       *      addresses.
       */
      DbgPrint("[+] TraceHook.SYS: Dumping the virtual memory of PID: 0x%.8x.\r\n",ProcessId);
      DumpProcessMemory(ProcessId);

      RemoveProcessFromList(ProcessId);
    }
  }
}



/* DriverEntry routine
 *
 * Arguments:
 *  Not important (refer to MSDN for details)
 * 
 * Return value:
 *  Not important (refer to MSDN for details)
 */
NTSTATUS
DriverEntry(PDRIVER_OBJECT DriverObject, PUNICODE_STRING RPath)
{  
  UNICODE_STRING u_DriverName, u_DosDeviceName;
  NTSTATUS NtStatus;
  UINT i;

  DbgPrint("[+] TraceHook.SYS: DriverEntry() called.\r\n");

  if(NT_SUCCESS(PsSetCreateProcessNotifyRoutine(NotifyRoutine,FALSE)))
  {
    DbgPrint("[+] TraceHook.SYS: Create notify-routine successfully set.\r\n");
    DbgPrint("[+] TraceHook.SYS: Registering an Unload routine.\r\n");
  	DriverObject->DriverUnload = DriverUnload;

    NameOffset = GetLocationOfProcessName();
    DbgPrint("[+] TraceHook.SYS: The EPROCESS name offset is: 0x%.4x\r\n",NameOffset);

    DbgPrint("[+] TraceHook.SYS: Creating a named device for the driver.\r\n");
   	RtlInitUnicodeString(&u_DriverName, DriverName);
  	RtlInitUnicodeString(&u_DosDeviceName, DosDeviceName);
    
    /* Create a named device in order to be able to communicate ring3<->ring0
     */
    NtStatus = IoCreateDevice(
			DriverObject,
			0,
			&u_DriverName,
			FILE_DEVICE_UNKNOWN,
			FILE_DEVICE_SECURE_OPEN,
			FALSE,
			&pDeviceObject
			);
	  if(!NT_SUCCESS(NtStatus))
		  return NtStatus;

    /* Create a symbolic link between the long and shortened driver name
     */
  	IoCreateSymbolicLink(
	  		&u_DosDeviceName,
		  	&u_DriverName
			  );

    /* Register the appropriate handler functions for the desired IRP events
     * (only interested in IOCTL in this case).
     */
  	for(i=0;i<IRP_MJ_MAXIMUM_FUNCTION;i++)
	  {
		  DriverObject->MajorFunction[i] = DriverUnsupported;
  	}
  	DriverObject->MajorFunction[IRP_MJ_DEVICE_CONTROL] = DriverCmd;    
    
    /* Initialize the ProcessId table
     */
    DbgPrint("[+] TraceHook.SYS: Allocating initial table of 4096 DWORD elements.\r\n");
    ProcessList.Count = 0;
    ProcessList.Size  = 4096;
    ProcessList.List  = ExAllocatePool(NonPagedPool,4096*sizeof(HANDLE));

    /* Initialize the dumping settings
     */
    DumpSettings.DumpType   = FullMemoryDump;
    DumpSettings.hThreadApc = NULL;
    DumpSettings.ApcRoutine = NULL;

    /* Initialize the spin-lock mechanism used further in this code
     */
    DbgPrint("[+] TraceHook.SYS: Initializing synchronization spin-locks.\r\n");
    KeInitializeSpinLock(&SpLock);
  }
  else
  {
    DbgPrint("[-] TraceHook.SYS: Failed to set a process notify callback.\r\n");
    return STATUS_UNSUCCESSFUL;
  }

  return STATUS_SUCCESS;
}


/* DriverUnload routine
 *
 * Arguments:
 *  Not important (refer to MSDN for details)
 * 
 * Return value:
 *  None
 *
 *  The function is responsible for removing the old process notify callback
 *  (not doing it would likely lead to a system crash due to an invalid pointer
 *  derefecence).
 */
VOID
DriverUnload(PDRIVER_OBJECT pDriverObject)
{
	UNICODE_STRING u_DosDeviceName;
  DbgPrint("[+] TraceHook.SYS: DriverUnload() called.\r\n");

  /* Ensure that no memory leak takes place
   */
  if(ProcessList.List)
  {
    ExFreePool(ProcessList.List);
    ProcessList.List = NULL;
  }

  if(NT_SUCCESS(PsSetCreateProcessNotifyRoutine(NotifyRoutine,TRUE)))
    DbgPrint("[+] TraceHook.SYS: Create notify routine removed.\r\n");
  else
    DbgPrint("[+] TraceHook.SYS: Failed to remove the callback routine\r\n");

  /* Unmap the symbolic link and delete the named device 
   */
	RtlInitUnicodeString(&u_DosDeviceName, DosDeviceName);
	IoDeleteSymbolicLink(&u_DosDeviceName);
	IoDeleteDevice(pDeviceObject);
}



/* DriverUnsupported routine
 *
 * Arguments:
 *  Not important (refer to MSDN for details)
 * 
 * Return value:
 *  Always returns STATUS_NOT_SUPPORTED.
 */
NTSTATUS 
DriverUnsupported(PDEVICE_OBJECT DeviceObject, PIRP Irp)
{
	return STATUS_NOT_SUPPORTED;
}


/* DriverCmd routine
 *
 * Arguments:
 *  Not important (refer to MSDN for details)
 * 
 * Return value:
 *  The value returned by separate signal-handling functions, indicating
 *  the execution state.
 *
 *  This routine is called everytime an IOCTL message is sent by a ring3 application.
 *  It's purpose is to analyze the signal and perform proper actions (call the right
 *  handler functions here) depending on the signal ID being passed.
 */
NTSTATUS 
DriverCmd(PDEVICE_OBJECT DeviceObject, PIRP Irp)
{
  NTSTATUS NtStatus = STATUS_SUCCESS;
	PIO_STACK_LOCATION irpStack;
	PVOID InBuffer, OutBuffer;
	ULONG InBufferSize, OutBufferSize;
	ULONG IOCtlCode;
	IO_STATUS_BLOCK IoStatusBlock;


  /* Security check - do NOT allow any process on the list communicate
   *                  with the TraceHook driver
   */
  if(IsProcessInList(PsGetCurrentProcessId())==TRUE)
  {
    NtStatus = STATUS_ACCESS_DENIED;
    goto ReturnStatus;
  }

	irpStack = IoGetCurrentIrpStackLocation(Irp);
	
  /* Get the basic data pointers provided by a user-mode application */
	InBuffer      = irpStack->Parameters.DeviceIoControl.Type3InputBuffer;
	InBufferSize  = irpStack->Parameters.DeviceIoControl.InputBufferLength;
	OutBuffer     = Irp->UserBuffer;
	OutBufferSize = irpStack->Parameters.DeviceIoControl.OutputBufferLength;
	IOCtlCode     = irpStack->Parameters.DeviceIoControl.IoControlCode;

	/* At the begining, assume everything will be ok */
	NtStatus = STATUS_SUCCESS;

	switch(IOCtlCode)
	{
		case IOCTL_QUERY_VERSION:
      NtStatus = HandleQueryVersion(InBuffer,OutBuffer,InBufferSize,OutBufferSize);
      break;
    case IOCTL_FINISH:
      NtStatus = HandleFinish(InBuffer,OutBuffer,InBufferSize,OutBufferSize);
      break;
    case IOCTL_ADD_PROCESS:
      NtStatus = HandleAddProcess(InBuffer,OutBuffer,InBufferSize,OutBufferSize);
      break;
    case IOCTL_REMOVE_PROCESS:
      NtStatus = HandleRemoveProcess(InBuffer,OutBuffer,InBufferSize,OutBufferSize);
      break;
    case IOCTL_SET_DUMP_PATH:
      NtStatus = HandleSetDumpPath(InBuffer,OutBuffer,InBufferSize,OutBufferSize);
      break;
    case IOCTL_SET_DUMP_MODE:
      NtStatus = HandleSetDumpMode(InBuffer,OutBuffer,InBufferSize,OutBufferSize);
      break;
    default:
      NtStatus = STATUS_NOT_SUPPORTED;
	}
	
ReturnStatus:
	Irp->IoStatus.Status = NtStatus;
	IoCompleteRequest(Irp, IO_NO_INCREMENT);

	return NtStatus;
}


/* HandleQueryVersion routine
 *
 * Arguments:
 *  The arguments contain standard IOCTL information (input and output buffer
 *  together with their sizes).
 * 
 * Return value:
 *  STATUS_SUCCESS if the function succeeds
 *  STATUS_ACCESS_VIOLATION or any other error code in case of failure (when reading the
 *  user-specified buffer etc)
 */
NTSTATUS
HandleQueryVersion(
	PVOID InBuffer,     // UNUSED
	PVOID OutBuffer,    
	ULONG InBufferSize, // UNUSED
	ULONG OutBufferSize
)
{
  DbgPrint("[+] TraceHook.SYS: HandleQueryVersion() called.\r\n");

  __try
  {
    if(OutBufferSize<sizeof(DriverVersion))
    {
      ExRaiseStatus(STATUS_INVALID_BUFFER_SIZE);
    }

    /* Check if there's enought space in the OutBuffer user-mode memory... */
    ProbeForWrite(OutBuffer,sizeof(DriverVersion),sizeof(CHAR));

    /* Copy the version string to the destination buffer */
    RtlCopyMemory(OutBuffer,DriverVersion,sizeof(DriverVersion));
  }
  except(EXCEPTION_EXECUTE_HANDLER)
  {
    return GetExceptionCode();
  }

  return STATUS_SUCCESS;
}


/* HandleFinish routine
 *
 * Arguments:
 *  None
 * 
 * Return value:
 *  TODO
 */
NTSTATUS
HandleFinish(
	PVOID InBuffer,       // UNUSED
	PVOID OutBuffer,      // UNUSED
	ULONG InBufferSize,   // UNUSED
	ULONG OutBufferSize   // UNUSED
)
{
	UNICODE_STRING u_DosDeviceName;
  UINT i;

  DbgPrint("[+] TraceHook.SYS: HandleFinish() called.\r\n");

  /* Iterate thru the entire process list, terminating and deleting
   * all the process currently present in the trace list
   */
  for( i=0;i<ProcessList.Count;i++ )
  {
    HANDLE hProcess;
    CLIENT_ID ClientId = {ProcessList.List[i],0};
    OBJECT_ATTRIBUTES ObjectAttributes;

    /* Initialize the structure with NULLs
     */
    InitializeObjectAttributes(&ObjectAttributes,NULL,0,NULL,NULL);

    if(NT_SUCCESS(ZwOpenProcess(&hProcess,PROCESS_TERMINATE,&ObjectAttributes,&ClientId)))
      ZwTerminateProcess(hProcess,1);
  }

  /* Wait until the list of processes is empty (all NotifyRoutine calls are complete)
   */
  while(ProcessList.Count!=0)
    __asm nop

  /* Free the list memory and zero out the structure itself
   */
  ProcessList.Size = ProcessList.Count = 0;
  if(ProcessList.List)
  {
    ExFreePool(ProcessList.List);
    ProcessList.List = NULL;
  }


  DbgPrint("[+] TraceHook.SYS: Finish signal successfully handled.\r\n");
  return STATUS_SUCCESS;
}


/* HandleAddProcess routine
 *
 * Arguments:
 *  TODO
 * 
 * Return value:
 *  TODO
 */
NTSTATUS
HandleAddProcess(
	PVOID InBuffer,
	PVOID OutBuffer,      // UNUSED 
	ULONG InBufferSize,
	ULONG OutBufferSize   // UNUSED
)
{
  DbgPrint("[+] TraceHook.SYS: HandleAddProcess() called.\r\n");
  
  __try
  {
    /* The specified input size must be equal sizeof(HANDLE)
     */
    if(InBufferSize!=sizeof(HANDLE))
    {
      ExRaiseStatus(STATUS_INVALID_BUFFER_SIZE);
    }

    /* Check the virtual address validity
     */
    ProbeForRead(InBuffer,InBufferSize,sizeof(BYTE));
    
    /* XXX: Is there anything else to perform a sanity check on?
     */

    /* Add the desired PID to our list
     */
    AddProcessToList(*(HANDLE*)InBuffer);
  }
  __except(EXCEPTION_EXECUTE_HANDLER)
  {
    return GetExceptionCode();
  }
  
  return STATUS_SUCCESS;
}


/* HandleRemoveProcess routine
 *
 * Arguments:
 *  TODO
 * 
 * Return value:
 *  TODO
 *
 *  This function is basically the same as HandleAddProcess - the only
 *  difference is the final internal function being called - RemoveProcessFromList.
 */
NTSTATUS
HandleRemoveProcess(
	PVOID InBuffer,
	PVOID OutBuffer,      // UNUSED 
	ULONG InBufferSize,
	ULONG OutBufferSize   // UNUSED
)
{
  DbgPrint("[+] TraceHook.SYS: HandleRemoveProcess() called.\r\n");
  
  __try
  {
    /* The specified input size must be equal sizeof(HANDLE)
     */
    if(InBufferSize!=sizeof(HANDLE))
    {
      ExRaiseStatus(STATUS_INVALID_BUFFER_SIZE);
    }

    /* Check the virtual address validity
     */
    ProbeForRead(InBuffer,InBufferSize,sizeof(BYTE));
    
    /* XXX: Is there anything else to perform a sanity check on?
     */

    /* Remove the desired PID from our list
     */
    RemoveProcessFromList(*(HANDLE*)InBuffer);
  }
  __except(EXCEPTION_EXECUTE_HANDLER)
  {
    return GetExceptionCode();
  }
  
  return STATUS_SUCCESS;
}


/* HandleSetDumpPath routine
 *
 * Arguments:
 *  A single ASCIIZ (null terminated) ANSI-encoded string.
 * 
 * Return value:
 *  STATUS_SUCCESS                - if the function succeeds
 *  STATUS_ACCESS_VIOLATION       - if one or more parameters are wrong
 *  STATUS_INVALID_BUFFER_SIZE    - if the buffer size exceeds MAX_PATH
 *
 *  This function sets the global dumping path of the driver. If not called, the driver uses
 *  the default C:\dump path.
 *  It is strongly advised to set a valid path every time the driver is loaded.
 *  
 *  Note: The function assumes that the directory path passed in the parameter is valid. 
 *        Since the malware is restricted and unable to communicate with the driver, 
 *        we should be relatively safe here.
 */
NTSTATUS
HandleSetDumpPath(
	PVOID InBuffer,
	PVOID OutBuffer,      // UNUSED
	ULONG InBufferSize,
	ULONG OutBufferSize   // UNUSED
)
{
  DbgPrint("[+] TraceHook.SYS: HandleSetDumpPath() called.\r\n");

  __try
  {
    /* Check, if the buffer is not too long
     */
    if(InBufferSize>sizeof(GlobalDumpPath)-sizeof('\0'))
    {
      ExRaiseStatus(STATUS_INVALID_BUFFER_SIZE);
    }

    /* Check the virtual address validity
     */
    ProbeForRead(InBuffer,InBufferSize,sizeof(BYTE));

    /* Copy the buffer contents to our global buffer
     */
    RtlCopyMemory(GlobalDumpPath,InBuffer,InBufferSize);
    GlobalDumpPath[InBufferSize] = '\0';
  }
  __except(EXCEPTION_EXECUTE_HANDLER)
  {
    return GetExceptionCode();
  }

  return STATUS_SUCCESS;
}



NTSTATUS
HandleSetDumpMode(
	PVOID InBuffer,
	PVOID OutBuffer,      // UNUSED
	ULONG InBufferSize,
	ULONG OutBufferSize   // UNUSED
)
{
  PVOID ThreadObject;

  DbgPrint("[+] TraceHook.SYS: HandleSetDumpMode() called.\r\n");

  __try
  {
    /* We want to get a DumpMode structure in the input param, nothing else
     */
    if(InBufferSize!=sizeof(DumpMode))
    {
      ExRaiseStatus(STATUS_INVALID_BUFFER_SIZE);
    }

    /* Check the virtual address validity
     */
    ProbeForRead(InBuffer,InBufferSize,sizeof(BYTE));

    /* Check the hThreadApc handle in case we're dealing with a minidump
     */
    if(((DumpMode*)InBuffer)->DumpType==MiniDump)
    {
      if(!NT_SUCCESS(ObReferenceObjectByHandle(
                      ((DumpMode*)InBuffer)->hThreadApc,
                      THREAD_SET_CONTEXT,
                      NULL,
                      UserMode,
                      &ThreadObject,
                      NULL)))
      {
        ExRaiseStatus(STATUS_INVALID_HANDLE);
      }
    }

    /* Fair enough, try to copy the buffer contents to our global structure
     */
    RtlCopyMemory(&DumpSettings,InBuffer,sizeof(DumpMode));
  }
  __except(EXCEPTION_EXECUTE_HANDLER)
  {
    return GetExceptionCode();
  }
  
  return STATUS_SUCCESS;
}

